import copy
import datetime
import importlib
import sys
import time
import re

import dateutil
import numpy as np
import pandas as pd
import json

from crossnetwork_processing_functions import get_data_files, create_GHOST_stage_logfile
from GHOST_standards import get_standard_data, get_standard_metadata, country_ISO3166_to_name, standard_data_flag_name_to_data_flag_code, standard_QA_name_to_QA_code, standard_sampling_types
from geographic_functions import dms2dd, local_timezone_name
from network_file_standardise import NetworkFileStandardise
from network_parsers import ParseEANETFile
from standardisation_mapping_functions import map_network_classifications_to_standard_classifications, map_network_measurement_information_to_standard_measurement_information, network_data_flag_to_standard_data_flag_mapping
from time_functions import round_to_minute, datetime_to_fractional_minutes_from_reference, get_datafile_download_timestamp
import unit_converter


class StandardiseNetworkData:

    '''Convert one data provider original file for a specific station into a GHOST standard numpy station specific file'''

    def __init__(self, station_reference, run_details, parameter_details, network_details):

        #add passed variables to self
        self.station_reference = station_reference
        self.run_details = run_details
        self.parameter_details = parameter_details
        self.network_details = network_details

        self.run_details['log_file_str'] += '{} -- PARSING RAW DATA\n'.format(self.station_reference)


    def do_work(self):
        '''Function which does executes the functions of the class in a pipeline'''

        #gather data files of station reference to parse
        data_files = self.gather_data_files()

        #parse station reference data files
        for data_file in data_files:
            self.parse_file(data_file)

        #if stage has completed successfully, with some valid data, write a valid process log file
        if self.written_data:
            self.run_details['log_file_str'] += str((time.time() - self.run_details['stage_process_start_time'])/60.)
            create_GHOST_stage_logfile(0,self.run_details)
        #else, if no data has been written for station write an error log file to inform station is not continuing in the pipeline
        elif not self.written_data:
            self.run_details['log_file_str'] += 'NO DATA HAS BEEN WRITTEN FOR STATION'
            create_GHOST_stage_logfile(1,self.run_details)


    def gather_data_files(self):

        '''Function to gather data files to parse per station'''

        #get data files specific to station
        data_files = get_data_files(self.run_details, self.parameter_details, specific_station=self.station_reference)
    
        #sort files in date order
        data_file_years = [int(re.sub('\D', '', data_file.split('/')[-1].split('_')[-1])) for data_file in data_files]
        data_files = np.array([x for _,x in sorted(zip(data_file_years,data_files))]) 
        data_file_years = sorted(data_file_years)    

        #load external metadata into memory and cut for station 
        # sys.path.append(self.run_details['data_root']+'EANET/metadata/processed')
        # meta_load = importlib.import_module('EANET_META')
        # self.external_metadata = meta_load.EANET_meta

        #load external metadata into memory 
        with open(self.run_details['data_root']+'EANET/metadata/{}/processed/EANET_META.json'.format(self.run_details['ghost_version']), 'r', encoding='utf-8') as f:
            self.external_metadata = json.loads(f.read())


        #get standard measured unit quantity
        conv_obj = unit_converter.convert_units(self.parameter_details['standard_units'],self.parameter_details['standard_units'],1)
        self.measured_unit_quantity = conv_obj.output_represented_quantity

        #set flag to know when have written data for station    
        self.written_data = False

        return data_files


    def parse_file(self, data_file):

        '''Function to parse data file to GHOST standard numpy station specific file'''

        self.run_details['log_file_str'] += 'PROCESSING FILE: {}\n'.format(data_file) 

        #get what type of file processing, i.e. AT, FP or PS?
        current_file_type = data_file.split('_')[-1].split('.')[0][4:]

        #get current file year
        current_file_year = data_file.split('_')[-1][:4] 

        #handle specific cases where EANET stations move through time record, this is handled by modifying station references (referenced in external metadata)
        if (self.station_reference == 'IDA001') & (int(current_file_year) >= 2016):
            station_reference = 'IDA001A'
        elif (self.station_reference == 'LAA001') & (current_file_type == 'AT'):
            station_reference = 'LAA001AUTO'
        elif (self.station_reference == 'LAA001') & (current_file_type in ['FP','PS']):
            station_reference = 'LAA001FP'
        elif (self.station_reference == 'MNA001') & (current_file_type == 'AT') & (self.run_details['parameter'] == 'PM2.5'):
            station_reference = 'MNA001PM25'
        elif (self.station_reference == 'MNA001') & (current_file_type == 'AT'):
            station_reference = 'MNA001AUTO'
        elif (self.station_reference == 'MNA001') & (current_file_type in ['FP','PS']):
            station_reference = 'MNA001FP'
        else:
            station_reference = copy.deepcopy(self.station_reference)
     
        #get cut of external metadata for station
        try:
            external_metadata_cut = self.external_metadata[station_reference]
        except:
            print('No metadata for this station reference {}'.format(station_reference))

        #create site instances of metadata/data dictionaries
        site_data =     get_standard_data(self.parameter_details)
        site_metadata = get_standard_metadata(self.parameter_details)

        #create object for parsing EANET file, and read file
        parsing_obj = ParseEANETFile(data_file, self.run_details)
        parsing_obj.read_data()

        #check if have data for parameter, if not skip file
        if self.run_details['parameter'] not in list(parsing_obj.parsed_data.keys()):
            self.run_details['log_file_str'] += 'NO DATA IN FILE FOR PARAMETER: {}. SKIPPING.\n'.format(data_file)
            return

        #check if have some valid data, if not skip file
        if len(parsing_obj.parsed_data[self.run_details['parameter']]) == 0:
            self.run_details['log_file_str'] += 'NO DATA IN FILE: {}. SKIPPING.\n'.format(data_file)
            return

        #------------------------------------------------------------------------#
        #PARSE METADATA
        #------------------------------------------------------------------------#

        #GHOST version
        site_metadata['GHOST_version']['value'].append(self.run_details['ghost_version'])

        #initialise string which will store all GHOST process warnings as empty            
        site_metadata['process_warnings']['value'].append('')

        #write known metadata  
        site_metadata['network']['value'].append(self.run_details['network'])
        site_metadata['data_level']['value'].append('none')  
        site_metadata['contact_institution']['value'].append('EANET')      
        site_metadata['contact_email_address']['value'].append('eanetdata@acap.asia')    
        site_metadata['data_licence']['value'].append('Rights reserved to the Network Center for the EANET: https://monitoring.eanet.asia/document/public/index')
        site_metadata['ellipsoid']['value'].append('WGS 84') 
        site_metadata['horizontal_datum']['value'].append('World Geodetic System 1984')
        site_metadata['vertical_datum']['value'].append('tidal - mean sea level')
        site_metadata['projection']['value'].append('geographic') 

        # add rain sampler to site metadata from external metadata
        if 'PREC' in self.run_details['parameter']:
            if ('Rain_Sampler' in list(external_metadata_cut.keys())) and (len(external_metadata_cut['Rain_Sampler']['values']) > 0):
                if (external_metadata_cut['Rain_Sampler']['values'][-1].replace(' ', '') == '') or (external_metadata_cut['Rain_Sampler']['values'][-1].replace(' ', '') == 'Unknown') or (external_metadata_cut['Rain_Sampler']['values'][-1].replace(' ', '') == 'Local'):
                    pass
                else:
                    if 'ECO' in external_metadata_cut['Rain_Sampler']['values'][-1].strip():
                        rain_sampler = 'ECOTECH Model 200'
                    elif 'Ogasawara' in external_metadata_cut['Rain_Sampler']['values'][-1].strip(): 
                        rain_sampler = 'Ogasawara rain sampler' 
                    elif 'Aerochem Metrics' in external_metadata_cut['Rain_Sampler']['values'][-1].strip(): 
                        rain_sampler = 'Aerochem Metrics 301' 
                    else:
                        rain_sampler = external_metadata_cut['Rain_Sampler']['values'][-1].strip()
                    
                    site_metadata['primary_sampling_instrument_name']['value'].append(rain_sampler)
                    if 'instrument_manual_name' in list(standard_sampling_types['wet-only precipitation chemistry sampler']['instruments'][rain_sampler].keys()):
                        site_metadata['primary_sampling_instrument_manual_name']['value'].append(standard_sampling_types['wet-only precipitation chemistry sampler']['instruments'][rain_sampler]['instrument_manual_name'])
                    if 'instrument_further_details' in list(standard_sampling_types['wet-only precipitation chemistry sampler']['instruments'][rain_sampler].keys()):
                        site_metadata['primary_sampling_further_details']['value'].append(standard_sampling_types['wet-only precipitation chemistry sampler']['instruments'][rain_sampler]['instrument_further_details'])

                    #site_metadata, measurement_qa_flags, standard_method_abbreviation = map_network_measurement_information_to_standard_measurement_information(rain_sampler, site_metadata, self.run_details, self.parameter_details)
                    #site_metadata['primary_sampling_instrument_name']['value'].append(rain_sampler)
            else:
                pass
        # add sampling height
        #sampling height is stated in 'Automatic_Measurement_Manual.pdf' that it should be between 3.0 and 10.0m, assume 3.0m 
        else:
            site_metadata['sampling_height']['value'].append('3.0')

        #add reported measurement unit
        #convert raw reported units given to standard name
        if self.run_details['parameter'] == 'PREC_ACCUM':
            reported_units = 'mm'
        elif '_PREC' in self.run_details['parameter']:
            reported_units = 'umol l-1'
        else:
            reported_units = parsing_obj.units[self.run_details['parameter']]

        #skip file if no units given no parameter
        if reported_units == '':            
            self.run_details['log_file_str'] += 'NO GIVEN UNITS IN FILE FOR PARAMETER: {}. SKIPPING.\n'.format(data_file)
            return
        conv_obj = unit_converter.convert_units(reported_units, reported_units, 1, measured_species = self.parameter_details['chemical_formula'])
        site_metadata['measuring_instrument_reported_units']['value'].append(conv_obj.output_standard_units)

        #add external metadata
        # dms_longitude_split = external_metadata_cut['longitude'].split(',')
        # dms_latitude_split = external_metadata_cut['latitude'].split(',')
        # if len(dms_longitude_split) == 2:
        #     orientation = dms_longitude_split[1][-1]
        #     dms_longitude_split[1] = dms_longitude_split[1][:-1]
        #     dms_longitude_split.append('00{}'.format(orientation))
        # if len(dms_latitude_split) == 2:
        #     orientation = dms_latitude_split[1][-1]
        #     dms_latitude_split[1] = dms_latitude_split[1][:-1]
        #     dms_latitude_split.append('00{}'.format(orientation))

        site_metadata['longitude']['value'].append(dms2dd(external_metadata_cut['Longitude_degree']['values'][-1],external_metadata_cut['Longitude_minute']['values'][-1] , external_metadata_cut['Longitude_second']['values'][-1], 'E'))
        site_metadata['latitude']['value'].append(dms2dd(external_metadata_cut['Latitude_degree']['values'][-1], external_metadata_cut['Latitude_minute']['values'][-1], external_metadata_cut['Latitude_second']['values'][-1], 'N'))
        site_metadata['altitude']['value'].append(external_metadata_cut['Altitude']['values'][-1])
        site_metadata['station_name']['value'].append(external_metadata_cut['Site']['values'][-1])
        if 'comments' in list(external_metadata_cut.keys()):        
            site_metadata['network_miscellaneous_details']['value'].append(external_metadata_cut['Note']['values'][-1])

        #parse station classifications (from external metadata)
        native_classifications = {'area_classification':external_metadata_cut['Classification']['values'][-1]}
        site_metadata = map_network_classifications_to_standard_classifications(native_classifications, site_metadata, self.run_details)

        #process measurement process information (from external metadata/Automatic_Measurement_Manual.pdf)
        if current_file_type == 'AT':
            if self.run_details['parameter'] in list(external_metadata_cut['methods'].keys()):
                measurement_information = external_metadata_cut['methods'][self.run_details['parameter']]
            else:
                measurement_information = '{}_at_default'.format(self.run_details['parameter'].lower())
        elif current_file_type == 'FP':
            if 'FP' in list(external_metadata_cut['methods'].keys()):
                measurement_information = external_metadata_cut['methods']['FP']
            else:
                measurement_information = 'fp_default'
        elif current_file_type == 'PS':
            if 'PS' in list(external_metadata_cut['methods'].keys()):
                measurement_information = external_metadata_cut['methods']['PS']
            else:
                measurement_information = 'ps_default'
        elif current_file_type == 'WET':
            if '_PREC' in self.run_details['parameter']:
                measurement_information = '{}_default'.format(self.run_details['parameter'].lower())
            elif self.run_details['parameter'] == 'PREC_ACCUM':
                if parsing_obj.parsed_data['method_code'][-1] == 1: # changes per measurement, so changes in time, include somehow in dynamic metadata
                    measurement_information = 'rain_gauge'
                #elif parsing_obj.parsed_data['method_code'][-1] == 2:
                #    measurement_information = 'derived_from_sample'
                else:
                    measurement_information = 'unknown'

        site_metadata, measurement_qa_flags, standard_method_abbreviation = map_network_measurement_information_to_standard_measurement_information(measurement_information, site_metadata, self.run_details, self.parameter_details)
        #set measurement qa codes which are True as joined metadata str
        site_metadata['measurement_qa_flags'] = {'value':[], 'data_type':object, 'string_format':'lower_short'}
        measurement_qa_flag_codes = []
        for standard_flag_name in measurement_qa_flags.keys():
            if measurement_qa_flags[standard_flag_name]:
                measurement_qa_flag_codes.append(str(standard_QA_name_to_QA_code[standard_flag_name]))
        site_metadata['measurement_qa_flags']['value'].append(','.join(measurement_qa_flag_codes))

        #iterate through metadata variables, if any columns have not been appended to add a NaN
        for column_header in list(site_metadata.keys()):
            n_column = len(site_metadata[column_header]['value'])
            #if column is empty, then append NaN
            if n_column < 1:
                site_metadata[column_header]['value'].append(np.NaN)

        #set station_reference (provided network ID + standard measurement method abbreviation)
        site_metadata['station_reference']['value'][-1] = '{}_{}'.format(station_reference,standard_method_abbreviation) 

        #get data file download timestamp
        site_metadata['data_download_stamp']['value'][-1] = get_datafile_download_timestamp(data_file)

        #add data revision timestamp, set as 0001-01-01 00:00 for all files
        data_revision_timestamp = datetime_to_fractional_minutes_from_reference(datetime.datetime(1,1,1,0,0))
        site_metadata['data_revision_stamp']['value'][-1] = data_revision_timestamp

        #------------------------------------------------------------------------#
        #PARSE TIME
        #------------------------------------------------------------------------#

        #get local timezone offset from UTC
        local_timezone = local_timezone_name(longitude=float(site_metadata['longitude']['value'][-1]), latitude=float(site_metadata['latitude']['value'][-1]))
        tzinfo = dateutil.tz.gettz(local_timezone)        

        #convert timezone to UTC
        start_dt = [round_to_minute(dt.replace(tzinfo=tzinfo).astimezone(dateutil.tz.tzutc())).replace(tzinfo=None) for dt in parsing_obj.parsed_data['start_dt']]
        end_dt = [round_to_minute(dt.replace(tzinfo=tzinfo).astimezone(dateutil.tz.tzutc())).replace(tzinfo=None) for dt in parsing_obj.parsed_data['end_dt']]

        #convert each start measurement datetime to minutes from reference time
        frac_time = [datetime_to_fractional_minutes_from_reference(dt) for dt in start_dt]
            
        #get timedeltas between start and end of measurement window
        temporal_resolutions = []
        for start, end in zip(start_dt,end_dt):
            dt_window = end - start
            #convert hours to minutes
            dt_window_minutes = int(dt_window.total_seconds()/60.)
            temporal_resolutions.append(dt_window_minutes)  

        #save meta_update_stamp, set as first sorted time processed (as minutes relative to reference time)
        site_metadata['meta_update_stamp']['value'][-1] = sorted(frac_time)[0]


        #------------------------------------------------------------------------#
        #PARSE QA/FLAGS
        #------------------------------------------------------------------------#

        flags = []
        qa = []

        #get dictionary necessary for mapping of US NADP MDN file data flags to standard data flag codes
        data_flag_mapping_dictionary = network_data_flag_to_standard_data_flag_mapping['EANET']      

        # join flags in one list of lists
        parsed_flags = []
        for ii, parsed_flag in enumerate(parsing_obj.parsed_data['flag1']):
            for n in [1,2,3]:
                raw_flag_a = parsing_obj.parsed_data['flag'+str(n)][ii].replace(' ', '').replace(',','').replace('`', '').replace('.', '')
                if len(raw_flag_a) == 6: # 6 digits, split in two
                    raw_flag_a = raw_flag_a[0:3]
                    
                    parsing_obj.parsed_data['flag'+str(n)][ii] = raw_flag_a
            parsed_flags.append([parsing_obj.parsed_data['flag1'][ii], parsing_obj.parsed_data['flag2'][ii], parsing_obj.parsed_data['flag3'][ii]])


        #iterate through network flags in each measurement
        for parsed_flag in parsed_flags:

            #lists to append qa/data flags per measurement
            flags_per_measurement = []
            qa_per_measurement = []
            #set GHOST decreed validity/network decreed validity variables initially as 'V'
            ghost_decreed_validity = 'V'
            network_decreed_validity = 'V'

            #iterate through flags per measurement
            for single_flag in parsed_flag:

                if single_flag != '':
                    # map note flag and append
                    standard_data_flag_dict = data_flag_mapping_dictionary[single_flag]
                    #get mapped standard data flags
                    standard_data_flag_names = standard_data_flag_dict['standard_data_flag_name']
                    #iterate through standard data flags 
                    for standard_data_flag_ii, standard_data_flag_name in enumerate(standard_data_flag_names):
                        #check if standard data flag is empty (i.e. unmapped as non-useful)
                        if standard_data_flag_name != '': 
                            #get corresponding data flag code for standard data flag
                            flag_code = standard_data_flag_name_to_data_flag_code[standard_data_flag_name]
                            #append standard data flag (if not already appended)
                            if flag_code not in flags_per_measurement:
                                flags_per_measurement.append(flag_code)
                            #determine if GHOST/network decreed valid
                            if standard_data_flag_dict['GHOST_decreed_validity'][standard_data_flag_ii] == 'I':
                                ghost_decreed_validity = 'I'
                            if standard_data_flag_dict['network_decreed_validity'][standard_data_flag_ii] == 'I':
                                network_decreed_validity = 'I'

            #if ghost decreed validity, or network decreed validity is Invalid, then set QA flags
            if ghost_decreed_validity == 'I':
                qa_per_measurement.append(standard_QA_name_to_QA_code['Invalid Data Provider Flags - GHOST Decreed'])
            if network_decreed_validity == 'I':
                qa_per_measurement.append(standard_QA_name_to_QA_code['Invalid Data Provider Flags - Network Decreed'])
    
            #append qa_per_measurement, flags_per_measurement to lists to write out
            flags.append(flags_per_measurement)
            qa.append(qa_per_measurement)


        #------------------------------------------------------------------------#
        #PARSE MEASUREMENTS
        #------------------------------------------------------------------------#

        #measurement values may need converting (i.e. mass density to mixing ratio)

        #measurement temperature and pressure values are never provided by EANET so are always assumed to be the network standard pressure and temperatures
        #these are defined as for 293.15 K temperature, and 1013.25 hPa for pressure here: Automatic_Measurement_Manual.pdf
        #Precipitation concentrations need conversion from umol l-1 to mg l-1, but no temperature/pressure normalization -> set network provided volume standard pressure/temperature to NaN
        
        measurements = []
        #initialise qa as all having timezone doubt qa flag, as get timezone from longitude/latitude
        qa = [[standard_QA_name_to_QA_code['Timezone Doubt']] for i in range(len(frac_time))]

        if 'PREC' in self.run_details['parameter']:
            site_metadata['network_provided_volume_standard_temperature']['value'][-1] = np.NaN
            site_metadata['network_provided_volume_standard_pressure']['value'][-1] = np.NaN
            standard_network_gas_volume = False
            unknown_gas_volume = False
        else:
            site_metadata['network_provided_volume_standard_temperature']['value'][-1] = '293.15'
            site_metadata['network_provided_volume_standard_pressure']['value'][-1] = '1013.25'
            standard_network_gas_volume = True
            unknown_gas_volume = False

        #get standard reported units/quantity
        standard_reported_units = conv_obj.output_standard_units # umol l-1
        measuring_instrument_reported_unit_quantity = conv_obj.output_represented_quantity  

        #set flags for writing QA flags associated with assumptions in unit conversion
        unit_conversion_standard_network = False
        unit_conversion_educated_guess = False

        #check if reported standard units, match required standard units
        if standard_reported_units != site_data[self.parameter_details['bsc_parameter_name']]['units']:

            #mass density to mixing ratio?
            if (measuring_instrument_reported_unit_quantity == 'mass_density') & (self.measured_unit_quantity == 'mole_fraction'):
                #get conversion factor
                input_units = {'temperature':'K', 'pressure':'hPa', 'molar_mass':'kg mol-1', 'mass_density':standard_reported_units}
                #do conversion with network provided sample gas temperature/pressure values
                #if these are not set, then make educated guess (assume network standards)
                if (not pd.isnull(site_metadata['network_provided_volume_standard_temperature']['value'][-1])) & (not pd.isnull(site_metadata['network_provided_volume_standard_pressure']['value'][-1])):   
                    input_vals = {'temperature':site_metadata['network_provided_volume_standard_temperature']['value'][-1], 'pressure':site_metadata['network_provided_volume_standard_pressure']['value'][-1], 'molar_mass':unit_converter.get_molecular_mass(self.parameter_details['chemical_formula']), 'mass_density':1.0}
                    unit_conversion_standard_network = True
                else:
                    input_vals = {'temperature':'293.15', 'pressure':'1013.25', 'molar_mass':unit_converter.get_molecular_mass(self.parameter_details['chemical_formula']), 'mass_density':1.0}
                    unit_conversion_educated_guess = True
                conv_obj = unit_converter.convert_units(input_units,site_data[self.parameter_details['bsc_parameter_name']]['units'],input_vals,measured_species = self.parameter_details['chemical_formula'],conversion_input_quantity = 'mass_density')

            #mixing ratio to mass density?
            elif (measuring_instrument_reported_unit_quantity == 'mole_fraction') & (self.measured_unit_quantity == 'mass_density'):
                input_units = {'temperature':'K', 'pressure':'hPa', 'molar_mass':'kg mol-1', 'mole_fraction':standard_reported_units}     
                #do conversion with network provided sample gas temperature/pressure values
                #if these are not set, then make educated guess (assume network standards)
                if (not pd.isnull(site_metadata['network_provided_volume_standard_temperature']['value'][-1])) & (not pd.isnull(site_metadata['network_provided_volume_standard_pressure']['value'][-1])): 
                    input_vals = {'temperature':site_metadata['network_provided_volume_standard_temperature']['value'][-1], 'pressure':site_metadata['network_provided_volume_standard_pressure']['value'][-1], 'molar_mass':unit_converter.get_molecular_mass(self.parameter_details['chemical_formula']), 'mole_fraction':1.0}
                    unit_conversion_standard_network = True
                else:
                    input_vals = {'temperature':'293.15', 'pressure':'1013.25', 'molar_mass':unit_converter.get_molecular_mass(self.parameter_details['chemical_formula']), 'mole_fraction':1.0}
                    unit_conversion_educated_guess = True
                conv_obj = unit_converter.convert_units(input_units,site_data[self.parameter_details['bsc_parameter_name']]['units'],input_vals,measured_species = self.parameter_details['chemical_formula'],conversion_input_quantity = 'mole_fraction')

            elif (measuring_instrument_reported_unit_quantity == 'molarity') & (self.measured_unit_quantity == 'mass_density'):
                input_units = {'molar_mass':'kg mol-1', 'molarity':standard_reported_units}     
                #do conversion 
                input_vals = {'molar_mass':unit_converter.get_molecular_mass(self.parameter_details['chemical_formula']), 'molarity':1.0}
                conv_obj = unit_converter.convert_units(input_units,site_data[self.parameter_details['bsc_parameter_name']]['units'],input_vals,measured_species=self.parameter_details['chemical_formula'],conversion_input_quantity = 'molarity')

            #otherwise simple unit conversion, i.e. pptv --> ppbv, ng/m3 --> ug/m3
            else:
                conv_obj = unit_converter.convert_units(standard_reported_units,site_data[self.parameter_details['bsc_parameter_name']]['units'],1.0, measured_species = self.parameter_details['chemical_formula'])

            #process through each measurement and convert value using derived conversion factor
            for n, meas in enumerate(parsing_obj.parsed_data[self.run_details['parameter']]):
                #only do conversion if measurement value is above 0 and not infinite
                try:
                    float_meas = float(meas)
                except:
                    float_meas = np.NaN
                
                if (float_meas > 0.0) & (np.isfinite(float_meas)):
                    measurements.append(float_meas*conv_obj.conversion_factor)
                else:
                    measurements.append(float_meas)

                #append sample gas volume/unit_conversion assumption QA flags if necessary
                if standard_network_gas_volume:
                    qa[n].append(standard_QA_name_to_QA_code['Sample Gas Volume - Network Standard'])
                if unknown_gas_volume:
                    qa[n].append(standard_QA_name_to_QA_code['Sample Gas Volume - Unknown'])
                if unit_conversion_standard_network:
                    qa[n].append(standard_QA_name_to_QA_code['Unit Conversion - Network Standard Sample Gas Volume Assumption'])
                if unit_conversion_educated_guess:
                    qa[n].append(standard_QA_name_to_QA_code['Unit Conversion - Educated Guess Sample Gas Volume Assumption'])

        #otherwise do nothing to measured values
        else:

            for n, meas in enumerate(parsing_obj.parsed_data[self.run_details['parameter']]):
                try:
                    float_meas = float(meas)
                except:
                    float_meas = np.NaN
                measurements.append(float_meas)

        #keep only measurement windows that are >= 0 minute in temporal resolution, as well as all other associated variables across other fields
        valid_temporal_resolution_inds = [ii for ii, res in enumerate(temporal_resolutions) if res >= 0]
        if len(valid_temporal_resolution_inds) != len(frac_time):
            frac_time = [frac_time[x] for x in valid_temporal_resolution_inds]
            temporal_resolutions = [temporal_resolutions[x] for x in valid_temporal_resolution_inds]
            measurements = [measurements[x] for x in valid_temporal_resolution_inds]
            qa = [qa[x] for x in valid_temporal_resolution_inds]

        #----------------------------------------------------------------------------#
        #WRITE OUT DATA
        #----------------------------------------------------------------------------# 
        site_data['time']['value'] = frac_time
        site_data[self.parameter_details['bsc_parameter_name']]['value'] = measurements
        site_data['temporal_resolution']['value'] = temporal_resolutions   
        #set data flags as all un-flagged
        site_data['flag']['value'] = [[standard_data_flag_name_to_data_flag_code['Un-Flagged Data']] for i in range(len(site_data['time']['value']))]
        site_data['qa']['value'] = qa
        #iterate through data variables, if any columns have not been appended to, add a list of NaNs (equal to length of time array)
        for column_header in list(site_data.keys()):
            n_column = len(site_data[column_header]['value'])
            #if column is empty, then append NaN
            if n_column == 0:
                if site_data[column_header]['data_type'] != np.uint8:
                    site_data[column_header]['value'] = [np.NaN]*len(site_data['time']['value'])
                else:
                    site_data[column_header]['value'] = [255]*len(site_data['time']['value'])
        #------------------------------------------------------

        #standardise metadata/data formats and write out station data to numpy file
        write_obj = NetworkFileStandardise(site_metadata, site_data, self.run_details, self.parameter_details)

        #update flag to state that have written data for station (if have done so)
        if write_obj.written:
            self.written_data = True
        else:
            self.run_details['log_file_str'] += 'DATA ALL MISSING IN FILE: {}. SKIPPING.\n'.format(data_file)